// This example is from _Java Examples in a Nutshell_. (http://www.oreilly.com)
// Copyright (c) 1997 by David Flanagan
// This example is provided WITHOUT ANY WARRANTY either expressed or implied.
// You may study, use, modify, and distribute it for non-commercial purposes.
// For any commercial use, see http://www.davidflanagan.com/javaexamples

import com.sun.java.swing.*;       // Beta package name; will change
import com.sun.java.swing.border.*;
import java.awt.*;

/**
 * This class demonstrates various methods of presenting choices with the JFC.
 * It also demonstrates the use of "tool tips" and the Box container.
 **/
public class ChoicePanel extends JBorderedPane {
  /** This constructor creates and sets up the contents of the pane */
  public ChoicePanel() {
    // Set the title for this pane
    this.setBorder(new JTitledBorder(this, "Choices"));

    // Create a font for use below
    Font f = new Font("SansSerif", Font.PLAIN, 14);

    // Some fruits that the user will be asked to choose among
    String[] choices = { "Cantaloupe", "Blueberry", "Persimmon", 
                         "Kumquat", "Kiwi", "Papaya", "Guava"};

    // Extra information about the fruits, to be displayed as "tool tips"
    String[] tips = {"Fragrant and Juicy", "Tart and sweet", "Surprising!",
                     "What is it?", "Jade green and icy", "Ambrosia", 
                     "Tropical delight"};

    // A box is a convenient combination of a container and a layout manager.
    // It can display its sub-components in columns or rows.  We'll use this
    // one to display a bunch of mutually-exclusive choices
    Box buttonbox1 = Box.createVerticalBox();
    // A ButtonGroup enforces mutual exclusion among JRadioButtons
    ButtonGroup group = new ButtonGroup();
    // Loop through the choices and create a bunch of JRadioButton objects
    for(int i = 0; i < choices.length; i++) {
      JRadioButton b = new JRadioButton(choices[i]);  // Create button
      b.setFont(f);                                   // Set the font
      b.setToolTipText(tips[i]);                      // Set a tool tip
      group.add(b);                                   // Add to exclusion group
      buttonbox1.add(b);                              // Add to the box
    }

    // This box is used to display non-exclusive choices with JCheckbox
    Box buttonbox2 = Box.createVerticalBox();
    for(int i = 0; i < choices.length; i++) {   // Loop through choices
      JCheckbox b = new JCheckbox(choices[i]);  // Create a checkbox object
      b.setFont(f);                             // Set the font
      b.setToolTipText(tips[i]);                // Set a tool tip
      buttonbox2.add(b);                        // Add it to the box
    }

    // Now we create a box to display the above to boxes side-by-side
    // Adding "glue" to a box between the primary components spaces 
    // those components out evenly.
    Box box1 = Box.createHorizontalBox();     // Create the box
    box1.add(Box.createGlue());               // Add glue on left
    box1.add(buttonbox1);                     // Add radio button box
    box1.add(Box.createGlue());               // Add glue in the middle
    box1.add(buttonbox2);                     // Add checkbox button box
    box1.add(Box.createGlue());               // Add glue on the right

    // This is another horizontal box.  It will hold a JListBox and a 
    // JComboBox.  Eventually, it will be placed below box1, above.
    Box box2 = Box.createHorizontalBox();       // Create the box
    box2.add(Box.createGlue());                 // Add some glue, of course

    // Create a JListBox to display the choices
    JListBox listbox = new JListBox(choices);   // Automatically handles arrays
    listbox.setToolTipText("Choose a fruit");   // Specify a tool tip
    box2.add(listbox);                          // Add the list to the box
    box2.add(Box.createGlue());                 // And put some glue after it

    // Now create a JComboBox to display the choices
    JComboBox combo = new JComboBox();          // The combo box
    combo.setToolTipText("Don't you get it?");  // A tooltip
    combo.setMaximumRowCount(4);                // How many rows in drop down
    for(int i = 0; i<choices.length; i++)       // Loop through the choices
      combo.addPossibleValue(choices[i]);       //   adding each one 
    box2.add(combo);                            // Add combo box to box
    box2.add(Box.createGlue());                 // And add glue

    // This is the box that stacks box1 and box2 from above.  
    Box box3 = Box.createVerticalBox();
    box3.add(box1);                             // No glue here...
    box3.add(box2);

    // One final box to hold some ordinary pushbuttons
    Box buttonbox3 = Box.createHorizontalBox();
    JButton b1 = new JButton("Okay");              // Create buttons
    JButton b2 = new JButton("Cancel");
    JButton b3 = new JButton("Help");
    f = new Font("SansSerif", Font.BOLD, 12);      // A custom font
    b1.setFont(f); b2.setFont(f); b3.setFont(f);   // Set the font
    buttonbox3.add(Box.createGlue());
    buttonbox3.add(b1); buttonbox3.add(Box.createGlue());  // Add buttons and
    buttonbox3.add(b2); buttonbox3.add(Box.createGlue());  // glue
    buttonbox3.add(b3); buttonbox3.add(Box.createGlue());

    // Finally, specify a layout manager and some insets for this pane.
    // Then center box3 (the radiobuttons, checkboxes, list, and combo box)
    // and put the pushbuttons on the bottom.
    this.setLayout(new BorderLayout(15, 15));
    this.setInsets(new Insets(15,15,15,15));
    this.add(box3, "Center");
    this.add(buttonbox3, "South");
  }
}
